/****************************************************************************
 *  This file is part of PPMVC project                                      *
 *  Written and distributed to public domain by Przemyslaw Skibinski, 2003  *
 *  Contents: File buffer for fast access to data after encoding            *
 ****************************************************************************/

int mCurrentPos,mUnreadedBytes;
int mBufferStart,mBufferStartBound,mBufferEnd,mBufferEndBound;
int notUpdate,notCheckCount,lp;
FILE* mFile;
unsigned char* mBuffer;
extern int maxLeftMatchLen;
int* commonBeforeLen;
extern int i;

#define MAX_BUFFER_SIZE 5120000
#define MOVE_BUFFER_LEN 1024000

inline void Init()
{
	mBufferStart=0;
	mBufferEnd=0;
	mBuffer=NULL;
	mCurrentPos=0;
	commonBeforeLen=new int[lastPointers];
}

inline void Deinit()
{
	if (mBuffer!=NULL)
		free(mBuffer-3);
	delete(commonBeforeLen);
}

inline void ReadFile(FILE *file)
{
	mFile=file;
	fread(mBuffer,1,mBufferEnd,file);
	mUnreadedBytes-=mBufferEnd;
}

inline void SetBuffer(int fileLen)
{
	mBufferStart=0;
	mBufferStartBound=mBufferStart+maxLeftMatchLen;
	mCurrentPos=mBufferStart-1;
	mBufferEnd=min(fileLen,MAX_BUFFER_SIZE);
	if (mBufferEnd==fileLen) // whole file in buffer
		mUnreadedBytes=0;
	else
		mUnreadedBytes=fileLen;

	mBuffer=(unsigned char*)malloc(mBufferEnd+6)+3;
	// watchers, for Common() functions - not to exceed buffer
	mBuffer[-3]=255;
	mBuffer[-2]=127;
	mBuffer[-1]=0;
	mBuffer[mBufferEnd]=0;
	mBuffer[mBufferEnd+1]=127;
	mBuffer[mBufferEnd+2]=255;

	mBufferEnd+=mBufferStart;
	mBufferEndBound=mBufferEnd-maxRightMatchLen;
}


inline void MoveBuffer(bool compress)
{
	if (mUnreadedBytes>0)
	{
		memmove(mBuffer,mBuffer+MOVE_BUFFER_LEN,MAX_BUFFER_SIZE-MOVE_BUFFER_LEN);
		mBufferStart+=MOVE_BUFFER_LEN;
		mBufferStartBound=mBufferStart+maxLeftMatchLen;
		mBufferEnd+=min(mUnreadedBytes,MOVE_BUFFER_LEN);
		mBufferEndBound=mBufferEnd-maxRightMatchLen;
		if (compress)
			mUnreadedBytes-=fread(mBuffer+MAX_BUFFER_SIZE-MOVE_BUFFER_LEN,1,min(mUnreadedBytes,MOVE_BUFFER_LEN),mFile);
	}
}


inline void AddChar(unsigned char c)
{
	mCurrentPos++;
	if (mCurrentPos>=mBufferEndBound && notUpdate==0)
		MoveBuffer(FALSE);

	if (mCurrentPos>=mBufferStart && mCurrentPos<mBufferEnd)
		mBuffer[mCurrentPos-mBufferStart]=c;
}

inline unsigned char* GetPointer(int pos)
{
	if (pos>=mBufferStart && pos<mBufferEnd)
		return &mBuffer[pos-mBufferStart];

	printf("error GetPointer\n");
	return NULL;
}

inline int GetChar()
{
	mCurrentPos++;
	if (mCurrentPos>=mBufferEndBound && notUpdate==0)
	{
		MoveBuffer(TRUE);
		if (mCurrentPos>=mBufferEnd)
			return -1;
	}

	return mBuffer[mCurrentPos-mBufferStart];
}


inline int GetChar(int pos)
{
	if (pos<mBufferStart)
		return 0;

	return mBuffer[pos-mBufferStart];
}


inline int GetCurrentPos()
{
	return mCurrentPos;
}

inline void ChangeCurrentPos(int move)
{
	mCurrentPos+=move;
}

inline void Print(int start,int end)
{
	for (int i=start; i<=end; i++)
		printf("%c",(i==10 || i==13)?32:mBuffer[i-mBufferStart]);
	printf(" ");
}


inline int Common(int offset1,int offset2, int bound)
{

	offset1-=mBufferStart;
	offset2-=mBufferStart;

	if (offset1<0)
		return 0;

	lp=0;
	while (mBuffer[offset1+lp]==mBuffer[offset2+lp] && lp<bound)
		lp++;

	return lp; 
	
}



inline bool CommonBeforeReverse(int offset1,int offset2, int bound)
{
	offset1-=mBufferStart;
	offset2-=mBufferStart;
	lp=bound-1;


	while (mBuffer[offset1-lp]==mBuffer[offset2-lp] && lp>=0)
		lp--;

	return lp<0;
}



inline int CommonBefore(int offset1,int offset2, int bound)
{
	offset1-=mBufferStart;
	offset2-=mBufferStart;
	lp=0;


	while (mBuffer[offset1-lp]==mBuffer[offset2-lp] && lp<bound)
		lp++;

	return lp;
}


int maxLen;
int maxNo;

inline int GetMaxCommonBefore(int* pointers, int& len)
{
	maxLen=0;
	maxNo=0;

	for (i=0; i<lastPointers; i++)
	{
		if (pointers[i]==0)
			break;

		if (pointers[i]<=mBufferStartBound)
			continue;

		len=CommonBefore(mCurrentPos-maxLen,pointers[i]-maxLen,maxLeftMatchLen-maxLen);
		if (len>0 && CommonBeforeReverse(mCurrentPos,pointers[i],maxLen))
		{
			len+=maxLen;
			maxLen=len-1;
			maxNo=i; 
		}
	}

	len=maxLen+1;

	return pointers[maxNo];
}



