/****************************************************************************
 *  This file is part of PPMVC project                                      *
 *  Written and distributed to public domain by Przemyslaw Skibinski, 2003  *
 *  Contents: Memory manager for fast and effective allocation of small     *
 *            memory blocks                                                 *
 ****************************************************************************/

#define MAX_MEMORY_MB 256
#define MEMORY_BLOCK_MB 1

#define BLOCKS_COUNT (MAX_MEMORY_MB/MEMORY_BLOCK_MB)

unsigned char* memoryBlock[BLOCKS_COUNT];
int freeBlock;
int freeMemSize;
unsigned char* freeMemPtr;

inline void InitMemoryManager()
{
	for (int i=0; i<BLOCKS_COUNT; i++)
		memoryBlock[i]=NULL;

	freeBlock=0;
	freeMemSize=0;
	freeMemPtr=0;
}

inline void RestartMemoryManager()
{
	for (int i=0; i<BLOCKS_COUNT; i++)
	{
		if (memoryBlock[i]!=NULL)
			free(memoryBlock[i]);
	}

	InitMemoryManager();
}

inline void AllocMemoryBlock()
{	
	if (freeBlock<BLOCKS_COUNT)
	{
		memoryBlock[freeBlock]=(unsigned char*)calloc(MEMORY_BLOCK_MB*1024*1024,1);
		if (memoryBlock[freeBlock]!=NULL)
		{
			freeMemSize=MEMORY_BLOCK_MB*1024*1024;
			freeMemPtr=memoryBlock[freeBlock];
			freeBlock++;
		}
	}	

}

inline unsigned char* Alloc(int size)
{
	if (freeMemSize>=size)
	{
		freeMemSize-=size;
		freeMemPtr+=size;
		return freeMemPtr-size;
	}
	else
	{
		AllocMemoryBlock();
		if (freeMemSize>=size)
		{
			freeMemSize-=size;
			freeMemPtr+=size;
			return freeMemPtr-size;
		}
	}

	return NULL;
}	

void PrintMemoryUsage()
{
	printf("blocks=%d (%d) allMem=%d\n",freeBlock,freeBlock*MEMORY_BLOCK_MB*1024*1024,(freeBlock+1)*MEMORY_BLOCK_MB*1024*1024-freeMemSize);
}