/****************************************************************************
 *  This file is part of PPMd project                                       *
 *  Written and distributed to public domain by Dmitry Shkarin 1997,        *
 *  1999-2001                                                               *
 *  Contents: main routine                                                  *
 *  Comments: system & compiler dependent file                              *
 ****************************************************************************/
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "PPMd.h"
#include "PPMdType.h"

#define PPM_LOGFILE "ppmvc.log"
#define PPMVC_NAME "PPMVC v1.0"
#define BACKSLASH '\\'
#define USAGE_STR "Usage: PPMd <e|d> [switches] <FileName[s] | Wildcard[s]>\n"
static const char* pFName;
static DWORD StartFilePosition;
static BOOL EncodeFlag;
static clock_t StartClock;
static struct ARC_INFO { // CRC? Hmm, maybe in another times...
    DWORD signature,attrib;
    WORD  info,FNLen,time,date;
	DWORD fileLen;
} _PACK_ATTR ai;

#if defined(_WIN32_ENVIRONMENT_)
#include <conio.h>

inline void EnvSetNormAttr(const char* FName) { SetFileAttributes(FName,FILE_ATTRIBUTE_NORMAL); }
inline int                         EnvGetCh() { return getch(); }
inline void           EnvGetCWD(char* CurDir) { GetCurrentDirectory(260,CurDir); }
inline void EnvSetDateTimeAttr(const char* WrkStr)
{
    FILETIME ft;
    DosDateTimeToFileTime(ai.date,ai.time,&ft);
    HANDLE hndl=CreateFile(WrkStr,GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,0);
    SetFileTime(hndl,&ft,NULL,&ft);         CloseHandle(hndl);
    SetFileAttributes(WrkStr,ai.attrib);
}
struct ENV_FIND_RESULT: public WIN32_FIND_DATA {
    const char*  getFName() const { return cFileName; }
    void copyDateTimeAttr() const {
        ai.attrib=dwFileAttributes;
        FileTimeToDosDateTime(&ftLastWriteTime,&ai.date,&ai.time);
    }
};
struct ENV_FILE_FINDER {
    HANDLE hndl;
    ENV_FIND_RESULT Rslt;
    ENV_FIND_RESULT getResult() { return Rslt; }
    BOOL isFileValid() {
        return ((Rslt.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0);
    }
    BOOL findFirst(const char* Pattern) {
        return ((hndl=FindFirstFile(Pattern,&Rslt)) != INVALID_HANDLE_VALUE);
    }
    BOOL findNext() { return FindNextFile(hndl,&Rslt); }
    void findStop() { FindClose(hndl); }
};

#elif defined(_DOS32_ENVIRONMENT_)
#include <conio.h>
#include <dos.h>
#include <direct.h>
#if defined(__DJGPP__)
#include <unistd.h>
#include <crt0.h>
#undef BACKSLASH
#define BACKSLASH '/'
char **__crt0_glob_function (char *arg) { return 0; }
void   __crt0_load_environment_file (char *progname) { }
#endif /* defined(__DJGPP__) */

inline void EnvSetNormAttr(const char* FName) { _dos_setfileattr(FName,_A_NORMAL); }
inline int                         EnvGetCh() { return getch(); }
inline void           EnvGetCWD(char* CurDir) { getcwd(CurDir,260); }
inline void EnvSetDateTimeAttr(const char* WrkStr)
{
    FILE* fpOut = fopen(WrkStr,"a+b");
    _dos_setftime(fileno(fpOut),ai.date,ai.time);
    fclose(fpOut);
    _dos_setfileattr(WrkStr,ai.attrib);
}
struct ENV_FIND_RESULT: public find_t {
    const char*  getFName() const { return name; }
    void copyDateTimeAttr() const {
        ai.attrib=attrib;
        ai.time=wr_time;                    ai.date=wr_date;
    }
};
struct ENV_FILE_FINDER {
    ENV_FIND_RESULT Rslt;
    ENV_FIND_RESULT getResult() { return Rslt; }
    BOOL isFileValid() { return TRUE; }
    BOOL findFirst(const char* Pattern) {
        return (_dos_findfirst(Pattern,_A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_ARCH,&Rslt) == 0);
    }
    BOOL findNext() { return (_dos_findnext(&Rslt) == 0); }
    void findStop() { }
};

#elif defined(_POSIX_ENVIRONMENT_)
#define _POSIX_SOURCE
#undef BACKSLASH
#define BACKSLASH '/'
#include <sys/stat.h>
#include <time.h>
#include <utime.h>
#include <dirent.h>
#include <unistd.h>
#include <fnmatch.h>
#if defined(__DJGPP__)
#include <crt0.h>
char **__crt0_glob_function (char *arg) { return 0; }
void   __crt0_load_environment_file (char *progname) { }
#endif /* defined(__DJGPP__) */

inline void EnvSetNormAttr(const char* FName) { chmod(FName,S_IWUSR); }
inline int                         EnvGetCh() { return getchar(); }
inline void           EnvGetCWD(char* CurDir) { getcwd(CurDir,260); }
inline void EnvSetDateTimeAttr(const char* WrkStr)
{
    struct utimbuf t;
    t.actime=t.modtime=(ai.date << 16)+ai.time;
    utime(WrkStr,&t);                       chmod(WrkStr,ai.attrib);
}
struct ENV_FIND_RESULT {
    dirent de;
    struct stat st;
    const char*  getFName() const { return de.d_name; }
    void copyDateTimeAttr() const {
        ai.attrib=st.st_mode;
        ai.time=st.st_mtime & 0xFFFF;       ai.date=st.st_mtime >> 16;
    }
};
struct ENV_FILE_FINDER {
    const char* pPattern;
    DIR* dir;
    dirent* de;
    struct stat st;
    ENV_FIND_RESULT getResult() {
        ENV_FIND_RESULT Rslt;
        Rslt.de=*de;                        Rslt.st=st;
        return Rslt;
    }
    BOOL isFileValid() {
        return (fnmatch(pPattern,de->d_name,FNM_NOESCAPE) == 0 &&
                stat(de->d_name,&st) == 0 && (st.st_mode & S_IRUSR) != 0 &&
                st.st_nlink == 1);
    }
    BOOL findFirst(const char* Pattern) {
        pPattern=Pattern;
        return ((dir=opendir(".")) && (de=readdir(dir)) != NULL);
    }
    BOOL findNext() { return ((de=readdir(dir)) != NULL); }
    void findStop() { closedir(dir); }
};

#else /* _UNKNOWN_ENVIRONMENT_ */
#pragma message ("unknown environment:")
#pragma message ("    1. _fastcall and _stdcall keywords are disabled")
#pragma message ("    2. wildcards and file attributes are disabled")

#undef  USAGE_STR
#define USAGE_STR "Usage: PPMd <e|d> [switches] FileName[s]\n"
inline void     EnvSetNormAttr(const char* ) {}
inline int                        EnvGetCh() { return getchar(); }
inline void          EnvGetCWD(char* CurDir) { CurDir[0]=0; }
inline void EnvSetDateTimeAttr(const char* ) {}
struct ENV_FIND_RESULT {
    char FName[260];
    const char*  getFName() const { return FName; }
    void copyDateTimeAttr() const {}
};
struct ENV_FILE_FINDER {
    const char* pPattern;
    ENV_FIND_RESULT getResult() {
        ENV_FIND_RESULT Rslt;               strcpy(Rslt.FName,pPattern);
        return Rslt;
    }
    BOOL isFileValid() { return TRUE; }
    BOOL findFirst(const char* Pattern) {
        pPattern=Pattern;                   return TRUE;
    }
    BOOL         findNext() { return FALSE; }
    void findStop() {}
};
#endif /* defined(__WIN32_ENVIRONMENT) */

static const char* const MTxt[] = { "Can`t open file %s\n",
    "read/write error for files %s/%s\n", "Out of memory!\n",
    "User break\n", "unknown command: %s\n", "unknown switch: %s\n",
    "designed and written by Dmitry Shkarin <dmitry.shkarin@mtu-net.ru>\n"
    "variable-length contexts (VC) by Przemyslaw Skibinski <inikep@o2.pl>\n"
    USAGE_STR
    "Switches (for encoding only):\n"
    "\t-oN    - set model order to N - [2,%d], default: %d\n"
    "\t-mN    - use N MB memory - [1,256], default: %d\n"
    "\t-d     - delete file[s] after processing, default: disabled\n"
    "\t-y     - delete archive if already exists, default: disabled\n"
    "\t-fName - set output file name to Name\n"
    "\t-a     - log results to file " PPM_LOGFILE "\n"
    "\t-rN    - set method of model restoration at memory insufficiency:\n"
    "\t\t-r0 - restart model from scratch (default)\n"
    "\t\t-r1 - cut off model (slow)\n"
    "\t\t-r2 - freeze model (dangerous)\n"
	"\n"
    "Switches (for VC testing only, should be specified for encoding and decoding):\n"
    "\t-g     - group size [1,256], default value depends on order\n"
	"\t-h     - min left (before) match lenght [1,10240], default value depends on order\n"
	"\t-i     - max left (before) match lenght [1,10240], default value depends on order\n"
	"\t-j     - min right (after) match lenght [1,10240], default value depends on order\n"
	"\t-l     - check [1,64] last pointers, default value depends on order\n"
	"\t-b     - ignore min left match len (faster compression)\n"
	"\t-u     - turn off variable-length contexts (ordinary PPMII compression)\n"
};
template <class T>
inline T CLAMP(const T& X,const T& LoX,const T& HiX) { return (X >= LoX)?((X <= HiX)?(X):(HiX)):(LoX); }
template <class T>
inline void SWAP(T& t1,T& t2) { T tmp=t1; t1=t2; t2=tmp; }

static double avg=0.0;
static int avgCount=0;
static int avgTime=0;

void _STDCALL PrintInfo(_PPMD_FILE* DecodedFile,_PPMD_FILE* EncodedFile, int pos, int additionalMemUsed,bool fileEnd)
{
    char WrkStr[320];
    UINT NDec=pos;
    NDec += (NDec == 0);
    UINT NEnc=ftell(EncodedFile)-StartFilePosition+sizeof(ARC_INFO);
    UINT n1=(8U*NEnc)/NDec;
    UINT n2=(100U*(8U*NEnc-NDec*n1)+NDec/2U)/NDec;
    if (n2 == 100) { n1++;                  n2=0; }
    int RunTime=((clock()-StartClock) << 10)/int(CLK_TCK);
    UINT Speed=NDec/(RunTime+(RunTime == 0));
    UINT UsedMemory=GetUsedMemory() >> 10;
    UINT m1=UsedMemory >> 10;
    UINT m2=(10U*(UsedMemory-(m1 << 10))+(1 << 9)) >> 10;
    if (m2 == 10) { m1++;                   m2=0; }
    if ( !EncodeFlag )                      
		SWAP(NDec,NEnc);

	sprintf(WrkStr,"%-12s %8d ->%8d, %1d.%03d bpc, used %d+%dMB, speed %d KB/s (%.1f s)",
           pFName,NDec,NEnc,n1,n2,m1,additionalMemUsed,Speed,(float)RunTime/1000);
    printf("%-79.79s\r",WrkStr);

	if (logResults && fileEnd)
	{
		avgCount++;
		avgTime+=RunTime;
		avg+=((float)NEnc/(float)NDec)*8.0;

		FILE* file=fopen(PPM_LOGFILE,"ab");

		if (file!=NULL)
		{
			fprintf(file,"%-10s %.03f (%d->%d) used %d+%dMB, speed %d KB/s (%.1f s)\r\n",pFName,((float)NEnc/(float)NDec)*8.0,NDec,NEnc,m1,additionalMemUsed,Speed,(float)RunTime/1000);
			fclose(file);
		}
	}
}
static char* _STDCALL ChangeExtRare(const char* In,char* Out,const char* Ext)
{
    char* RetVal=Out;
    const char* p=strrchr(In,'.');
    if (!p || strrchr(In,BACKSLASH) > p)    p=In+strlen(In);
    do { *Out++ = *In++; } while (In != p);
    *Out++='.';
    while((*Out++ = *Ext++) != 0)           ;
    return RetVal;
}
inline BOOL RemoveFile(const char* FName)
{
    EnvSetNormAttr(FName);                  return (remove(FName) == 0);
}

static BOOL YesToAll=FALSE;
static BOOL _STDCALL TestAccessRare(const char* FName)
{
    FILE* fp=fopen(FName,"rb");
    if ( !fp )                              return TRUE;
    fclose(fp);
    if ( YesToAll )                         return RemoveFile(FName);
    printf("%s already exists, overwrite?: <Y>es, <N>o, <A>ll, <Q>uit?",FName);
    for ( ; ; )
        switch ( toupper(EnvGetCh()) ) {
            case 'A':                       YesToAll=TRUE;
            case '\r': case 'Y':            return RemoveFile(FName);
            case 0x1B: case 'Q':            printf(MTxt[3]); exit(-1);
            case 'N':                       return FALSE;
        }
}
inline FILE* FOpen(const char* FName,const char* mode)
{
    FILE* fp=fopen(FName,mode);
    if ( !fp ) { printf(MTxt[0],FName);     exit(-1); }
    setvbuf(fp,NULL,_IOFBF,64*1024);        return fp;
}
inline void PrepareCoding(int SASize,FILE* fp)
{
    if ( !StartSubAllocator(SASize) ) {
        printf(MTxt[2]);                    exit(-1);
    }
    StartClock=clock();                     StartFilePosition=ftell(fp);
}

long filesize(FILE *stream) {
	long curpos, length;
	curpos = ftell(stream);
	fseek(stream, 0L, SEEK_END);
	length = ftell(stream);
	fseek(stream, curpos, SEEK_SET);
	return length;
}

void InitParams(int MaxOrder,int SASize);

inline void EncodeFile(const ENV_FIND_RESULT& efr,int MaxOrder,int SASize,MR_METHOD MRMethod,const char* ArcName)
{
    char WrkStr[260];
    strcpy(WrkStr,ArcName);
    if (!WrkStr[0] && !TestAccessRare(ChangeExtRare(efr.getFName(),WrkStr,"pmv")))
                return;
    FILE* fpIn = FOpen(efr.getFName(),"rb"), * fpOut = FOpen(WrkStr,"a+b");
    pFName=strrchr(efr.getFName(),BACKSLASH);
    pFName=( pFName )?(pFName+1):(efr.getFName());
    efr.copyDateTimeAttr();
    ai.signature=PPMdSignature;             
	ai.FNLen=strlen(pFName)+(MRMethod << 14);
    ai.info=(MaxOrder-1) | ((SASize-1) << 4) | ((Variant-'A') << 12);
	ai.fileLen=filesize(fpIn);
    fwrite(&ai,sizeof(ai),1,fpOut);         
	fwrite(pFName,ai.FNLen & 0x1FF,1,fpOut);
    PrepareCoding(SASize,fpOut);       
	InitParams(MaxOrder,SASize);
	EncodeFile(fpOut,fpIn,MaxOrder,MRMethod,ai.fileLen);
    putchar('\n');
    if (ferror(fpOut) || ferror(fpIn)) {
        printf(MTxt[1],efr.getFName(),WrkStr);
        exit(-1);
    }
    fclose(fpIn);                           
	fclose(fpOut);
}

inline BOOL DecodeOneFile(FILE* fpIn)
{
    char WrkStr[260];
    int MaxOrder, SASize;
    MR_METHOD MRMethod;
    if ( !fread(&ai,sizeof(ai),1,fpIn) )    
		return FALSE;
    MRMethod=MR_METHOD(ai.FNLen >> 14);
    ai.FNLen=CLAMP(int(ai.FNLen & 0x1FF),1,260-1);
    fread(WrkStr,ai.FNLen,1,fpIn);          
	WrkStr[ai.FNLen]=0;
    if ( !TestAccessRare(WrkStr) )          
		return FALSE;
    FILE* fpOut = FOpen(pFName=WrkStr,"wb");
    MaxOrder=(ai.info & 0x0F)+1;            
	SASize=((ai.info >> 4) & 0xFF)+1;
 	InitParams(MaxOrder,SASize);
	DWORD Variant=(ai.info >> 12)+'A';
    if (ai.signature != PPMdSignature || Variant != ::Variant) 
	{
        printf(MTxt[0],WrkStr);
		exit(-1);
    }
    PrepareCoding(SASize,fpIn);             
	DecodeFile(fpOut,fpIn,MaxOrder,MRMethod,ai.fileLen);
    putchar('\n');
    if (ferror(fpOut) || ferror(fpIn) || feof(fpIn)) 
	{
        printf(MTxt[1],WrkStr,WrkStr);     
		exit(-1);
    }
    fclose(fpOut);                          EnvSetDateTimeAttr(WrkStr);
    return TRUE;
}
inline void DecodeFile(const ENV_FIND_RESULT& efr)
{
    FILE* fpIn=FOpen(efr.getFName(),"rb");
    while ( DecodeOneFile(fpIn) )           ;
    fclose(fpIn);
}
inline void TestArchive(char* ArcName,const char* Pattern)
{
    if ( !Pattern[0] ) {
        char CurDir[260];
        EnvGetCWD(CurDir);
        const char* p=strrchr(CurDir,BACKSLASH);
        p = (p && strlen(p+1))?(p+1):("PPMdFile");
        ChangeExtRare(p,ArcName,"pmv");
    } else                                  strcpy(ArcName,Pattern);
    FILE* fp = fopen(ArcName,"rb");
    if ( fp ) {
        if (!fread(&ai,sizeof(ai),1,fp) || ai.signature != PPMdSignature ||
                                            (ai.info >> 12)+'A' != ::Variant) {
            printf(MTxt[0],ArcName);        exit(-1);
        }
        fclose(fp);
    }
}
struct FILE_LIST_NODE {
    FILE_LIST_NODE* next;
    ENV_FIND_RESULT efr;
    FILE_LIST_NODE(const ENV_FIND_RESULT& Data,FILE_LIST_NODE** PrevNode) {
        efr=Data;                           next=*PrevNode;
        *PrevNode=this;
    }
    void destroy(FILE_LIST_NODE** PrevNode) {
        *PrevNode=next;                     delete this;
    }
};

bool printProbability; // flag for testing
bool unboundedContexts,ignoreLeftMatch;
int minLeftMatchLen,minRightMatchLen,maxLeftMatchLen,maxRightMatchLen;
int groupMatches,notCheck,lastPointers,logResults;

void InitParams(int MaxOrder,int SASize)
{
	if (notCheck==0)
	{

		if (MaxOrder<=4)
			notCheck=4;
		else
		if (MaxOrder<=8)
			notCheck=6;
		else
		if (MaxOrder<=16)
			notCheck=8;
	}

	if (lastPointers==0)
	{

		if (MaxOrder<=4)
			lastPointers=8;
		else
		if (MaxOrder<=8)
			lastPointers=4;
		else
		if (MaxOrder<=16)
			lastPointers=2;
	}

	if (ignoreLeftMatch)
		lastPointers=1;

	if (groupMatches==0)
		groupMatches=MaxOrder+1;

	if (minRightMatchLen==0)
		minRightMatchLen=groupMatches*2;

	if (minLeftMatchLen==0)
	{
		minLeftMatchLen=MaxOrder;
		switch (MaxOrder)
		{
			case 4:
				minLeftMatchLen=6;
				break;
			case 5:
				minLeftMatchLen=8;
				break;
			case 6:
				minLeftMatchLen=10;
				break;
			case 7:
				minLeftMatchLen=12;
				break;
		}
		if (MaxOrder>=8)
			minLeftMatchLen=(MaxOrder-1)*2;
	}

	if (maxLeftMatchLen==0)
		maxLeftMatchLen=255;

	if (maxRightMatchLen==0)
		maxRightMatchLen=10240; 

	if (logResults==1)
	{
		FILE* file=fopen(PPM_LOGFILE,"ab");
		if (file!=NULL)
		{
			fprintf(file,"\r\n" PPMVC_NAME "  o=%d m=%d %c=%d g=%d(Group) h=%d(LMinML) i=%d(LMaxML) j=%d(RMinML)\r\n",MaxOrder,SASize,!unboundedContexts?'u':(ignoreLeftMatch?'b':'l'),lastPointers,groupMatches,minLeftMatchLen,maxLeftMatchLen,minRightMatchLen);
			fclose(file);
		}
		logResults++;
	}
}

int main(int argc, char *argv[])
{
    char ArcName[260];
    BOOL DeleteFile=FALSE;
    int i, MaxOrder=8, SASize=32;
	printProbability=FALSE;
	logResults=0;
	unboundedContexts=TRUE;
	ignoreLeftMatch=FALSE;
	notCheck=lastPointers=maxLeftMatchLen=maxRightMatchLen=minLeftMatchLen=minRightMatchLen=groupMatches=0;

	MR_METHOD MRMethod=MRM_RESTART;
    printf(PPMVC_NAME " ("__DATE__") - Fast PPMII+VC compressor\n");
    if (argc < 3) { printf(MTxt[6],SASize,MAX_O,MaxOrder);      return -1; }
    switch ( toupper(argv[1][0]) ) {
        case 'E': EncodeFlag=TRUE;                              break;
        case 'D': EncodeFlag=FALSE;                             break;
        default : printf(MTxt[4],argv[1]);                      return -1;
    }
    for (ArcName[0]=0,i=2;i < argc && (argv[i][0] == '-' || argv[i][0] == '/');i++)
        switch ( toupper(argv[i][1]) ) {
            case 'Y': YesToAll=TRUE;                            break;
            case 'D': DeleteFile=TRUE;                          break;
            case 'F': TestArchive(ArcName,argv[i]+2);           break;
            case 'M': SASize=CLAMP(atoi(argv[i]+2),1,256);      break;
            case 'O': MaxOrder=CLAMP(atoi(argv[i]+2),2,MAX_O);  break;
            case 'P': printProbability=TRUE;					break;
            case 'U': unboundedContexts=FALSE;					break;
            case 'A': logResults=1;								break;
            case 'B': ignoreLeftMatch=TRUE;						break;
            case 'R': MRMethod=MR_METHOD(CLAMP(atoi(argv[i]+2),0,2)); break;
            case 'G': groupMatches=CLAMP(atoi(argv[i]+2),1,256); break;
            case 'H': minLeftMatchLen=CLAMP(atoi(argv[i]+2),1,10240); break;
            case 'I': maxLeftMatchLen=CLAMP(atoi(argv[i]+2),1,10240); break;
            case 'J': minRightMatchLen=CLAMP(atoi(argv[i]+2),1,10240); break;
            case 'L': lastPointers=CLAMP(atoi(argv[i]+2),1,4096); break;
            default : printf(MTxt[5],argv[i]);   				return -1;
        }

    FILE_LIST_NODE* pNode, * pFirstNode=NULL, ** ppNode=&pFirstNode;
    for (ENV_FILE_FINDER eff;i < argc;i++) {
        if ( eff.findFirst(argv[i]) )
            do {
                if ( eff.isFileValid() ) {
                    pNode = new FILE_LIST_NODE(eff.getResult(),ppNode);
                    if ( !pNode ) {
                        printf(MTxt[2]);    return -1;
                    }
                    ppNode=&(pNode->next);
                }
            } while ( eff.findNext() );
        eff.findStop();
    }


    while ((pNode=pFirstNode) != NULL) 
	{
		ENV_FIND_RESULT& efr=pNode->efr;
		if ( EncodeFlag )
			EncodeFile(efr,MaxOrder,SASize,MRMethod,ArcName);
		else                                
			DecodeFile(efr);
	
		if ( DeleteFile )                   
			remove(efr.getFName());
        pNode->destroy(&pFirstNode);
    }

	if (logResults)
	{
		FILE* file=fopen(PPM_LOGFILE,"ab");
		if (file!=NULL)
		{
			fprintf(file,"%-8s %.04f, time=%.2f s\r\n","average",avg/avgCount,(float)avgTime/1000);
			fclose(file);
		}
	}

    StopSubAllocator();
    return 0;
}
